import json
from pathlib import Path
from uuid import uuid4
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from ..common import (
    QueryParser, QueryDomain, TenantCatalogError, TenantRetrievalError)
from ..models import Tenant
from .cataloguer import Cataloguer


class JsonCataloguer(Cataloguer):

    def __init__(self, path: str, parser: QueryParser) -> None:
        self.path = path
        self.parser = parser
        self.collection = 'tenants'
        self.catalog_schema: Dict = {
            self.collection: {}
        }
        self._setup()

    def _setup(self) -> None:
        catalog_file = Path(self.path)

        if not catalog_file.exists():
            with catalog_file.open('w') as f:
                json.dump(self.catalog_schema, f, indent=2)
            return

        with catalog_file.open('r+') as f:
            try:
                data = json.load(f)
                if self.collection in data:
                    return
            except json.JSONDecodeError as e:
                pass

            json.dump(self.catalog_schema, f, indent=2)

    def add_tenant(self, tenant: Tenant) -> Tenant:
        path = Path(self.path)
        with path.open() as f:
            try:
                data = json.load(f)
            except json.JSONDecodeError as e:
                raise TenantCatalogError(
                    f"The tenant catalog file <${self.path}> "
                    "is not valid json.")

        tenant.id = tenant.id or str(uuid4())

        data[self.collection].update({tenant.id: vars(tenant)})
        with path.open('w') as f:
            json.dump(data, f, indent=2)

        return tenant

    def get_tenant(self, tenant_id: str) -> Tenant:
        with open(self.path) as f:
            data = json.load(f)
            tenants = data.get(self.collection, {})
            tenant_dict = tenants.get(tenant_id)
            if not tenant_dict:
                raise TenantRetrievalError(
                    f"The entity with id {tenant_id} was not found.")
            return Tenant(**tenant_dict)

    def search_tenants(self, domain: QueryDomain) -> List[Tenant]:
        with open(self.path, 'r') as f:
            data = json.load(f)
            tenants_dict = data.get(self.collection, {})

        tenants = []
        filter_function = self.parser.parse(domain)
        for tenant_dict in tenants_dict.values():
            tenant = Tenant(**tenant_dict)

            if filter_function(tenant):
                tenants.append(tenant)

        return tenants



from abc import ABC, abstractmethod
from uuid import uuid4
from typing import List, Dict, Any, Optional
from ..models import Tenant


class Provisioner(ABC):
    """Tenant Provision service."""

    @abstractmethod
    def provision_tenant(self, tenant: Tenant) -> None:
        "Provision tenant method to be implemented."



from abc import ABC, abstractmethod
from typing import Dict, Optional
from pathlib import Path
from shutil import copytree
from uuid import uuid4
from ..common import TenantProvisionError
from ..models import Tenant
from .provisioner import Provisioner


class DirectoryProvisioner(Provisioner):

    def __init__(self, template: str, data: str) -> None:
        self.template = template
        self.data = data

    def provision_tenant(self, tenant: Tenant) -> None:
        tenant_directory = str(Path(self.data) / tenant.slug)
        copytree(self.template, tenant_directory)



from typing import Dict, List, Optional, Any
from .models import Tenant
from .cataloguer import Cataloguer
from .provisioner import Provisioner
from .common import QueryDomain, TenantCreationError


class Arranger:
    def __init__(self, cataloguer: Cataloguer,
                 provisioner: Provisioner) -> None:
        self.cataloguer = cataloguer
        self.provisioner = provisioner

    def create_tenant(self, tenant_dict: Dict[str, Any]) -> None:
        tenant = Tenant(**tenant_dict)
        domain: QueryDomain = ['|', ('slug', '=', tenant.slug),
                               ('name', '=', tenant.name)]
        duplicates = self.cataloguer.search_tenants(domain)

        if duplicates:
            raise TenantCreationError(
                f'A tenant with slug "{tenant.slug}" already exists.')

        tenant = self.cataloguer.add_tenant(tenant)
        self.provisioner.provision_tenant(tenant)



import time
import unicodedata
from typing import Mapping
from ..common import TenantLocationError, TenantCreationError


class Tenant:
    def __init__(self, **attributes):
        now = int(time.time())
        self.id = attributes.get('id', '')
        self.created_at = now
        self.updated_at = now
        self.name = attributes['name']
        self.email = attributes.get('email', '')
        self.active = attributes.get('active', True)
        self.slug = self._normalize_slug(attributes.get('slug', self.name))
        self.attributes = attributes.get('attributes', {})
        self.data: Mapping[str, str] = attributes.get('data', {
            'memory': self.slug
        })

    def location(self, type: str = 'memory') -> str:
        if type not in self.data:
            raise TenantLocationError(
                f"No location found for '{type}' type "
                f"in tenant '{self.name}'.")
        return self.data[type]

    @staticmethod
    def _normalize_slug(slug: str) -> str:
        stripped_slug = slug.strip().replace(" ", "_").lower()
        normalized_slug = unicodedata.normalize(
            'NFKD', stripped_slug).encode('ascii', 'ignore').decode('utf-8')
        if not normalized_slug:
            raise TenantCreationError("Invalid tenant 'slug' name.")
        return normalized_slug



from abc import ABC, abstractmethod
from typing import Dict, Optional
from uuid import uuid4
from ..common import TenantProvisionError
from ..models import Tenant
from .provisioner import Provisioner


class MemoryProvisioner(Provisioner):

    def __init__(self) -> None:
        self.pool: Dict[str, Tenant] = {}

    def provision_tenant(self, tenant: Tenant) -> None:
        self.pool[tenant.id] = tenant



from uuid import uuid4
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from ..common import QueryParser, QueryDomain, TenantRetrievalError
from ..models import Tenant
from .cataloguer import Cataloguer


class MemoryCataloguer(Cataloguer):

    def __init__(self, parser: QueryParser) -> None:
        self.parser = parser
        self.catalog: Dict[str, Tenant] = {}

    def add_tenant(self, tenant: Tenant) -> Tenant:
        tenant.id = tenant.id or str(uuid4())
        self.catalog[tenant.id] = tenant
        return tenant

    def get_tenant(self, tenant_id: str) -> Tenant:
        tenant = self.catalog.get(tenant_id)

        if not tenant:
            raise TenantRetrievalError('Tenant not found.')

        return tenant

    def search_tenants(self, domain: QueryDomain) -> List[Tenant]:
        tenants = []
        filter_function = self.parser.parse(domain)
        for tenant in self.catalog.values():
            if filter_function(tenant):
                tenants.append(tenant)

        return tenants



from typing import Dict, Optional
from .common import QueryParser
from .cataloguer import (
    Cataloguer, MemoryCataloguer, JsonCataloguer)
from .provisioner import (
    Provisioner, MemoryProvisioner, DirectoryProvisioner)
from .provider import Provider
from .arranger import Arranger


def resolve_cataloguer(options: Dict[str, str]) -> Cataloguer:
    parser = QueryParser()
    cataloguer: Optional[Cataloguer] = None
    if options.get('cataloguer_kind') == 'json':
        path = options['catalog_path']
        cataloguer = JsonCataloguer(path, parser)
    else:
        cataloguer = MemoryCataloguer(parser)

    return cataloguer


def resolve_provisioner(options: Dict[str, str]) -> Provisioner:
    parser = QueryParser()
    provisioner: Optional[Provisioner] = None
    if options.get('provisioner_kind') == 'directory':
        template = options['provision_template']
        data = options['data_directory']
        provisioner = DirectoryProvisioner(template, data)
    else:
        provisioner = MemoryProvisioner()

    return provisioner


def resolve_arranger(options: Dict[str, str]) -> Arranger:
    cataloguer = resolve_cataloguer(options)
    provisioner = resolve_provisioner(options)
    arranger = Arranger(cataloguer, provisioner)
    return arranger


def resolve_provider(options: Dict[str, str]) -> Provider:
    cataloguer = resolve_cataloguer(options)
    provider = Provider(cataloguer)
    return provider



from typing import Dict, Any
from .common import QueryDomain, TenantRetrievalError
from .models import Tenant
from .cataloguer import Cataloguer


class Provider:

    def __init__(self, cataloguer: Cataloguer) -> None:
        self.cataloguer = cataloguer

    def get_tenant(self, tenant_id: str) -> Dict[str, Any]:
        tenant = self.cataloguer.get_tenant(tenant_id)
        return vars(tenant)

    def resolve_tenant(self, name: str) -> Dict[str, Any]:
        domain: QueryDomain = ['|', ('slug', '=', name),
                               ('name', '=', name)]
        entities = self.cataloguer.search_tenants(domain)
        if not entities:
            raise TenantRetrievalError("Tenant not found.")

        tenant = entities.pop()
        return vars(tenant)



from uuid import uuid4
from abc import ABC, abstractmethod
from typing import List, Dict, Any, Optional
from ..common import QueryParser, QueryDomain
from ..models import Tenant


class Cataloguer(ABC):
    """Tenant Catalog Supplier."""

    @abstractmethod
    def add_tenant(self, tenant: Tenant) -> Tenant:
        "Add tenant method to be implemented."

    @abstractmethod
    def get_tenant(self, tenant_id: str) -> Tenant:
        "Get tenant method to be implemented."

    @abstractmethod
    def search_tenants(self, domain: QueryDomain) -> List[Tenant]:
        "Search tenants method to be implemented."




class TenarkError(Exception):
    """Tenark's base error class."""


# Tenancy

class TenantError(TenarkError):
    """Tenancy base error class."""


class TenantLocationError(TenantError):
    """The tenant location type was not found."""


class TenantCreationError(TenantError):
    """The tenant couldn't be created."""


class TenantRetrievalError(TenantError):
    """The tenant couldn't be found."""


class TenantProvisionError(TenantError):
    """The tenant couldn't be provisioned."""


class TenantCatalogError(TenantError):
    """The tenant couldn't be provisioned."""



from .arranger import Arranger
from .provider import Provider
from .resolver import resolve_provider, resolve_arranger


def build_provider(**options) -> Provider:
    provider = resolve_provider(options)
    return provider


def build_arranger(**options) -> Arranger:
    arranger = resolve_arranger(options)
    return arranger
